//
//  SwUart.c
//  fab-synth-clocknode
//
//  Created by James Fletcher on 09/06/2014.
//
//

#include <avr/io.h>
#include <avr/cpufunc.h>
#include <avr/interrupt.h>
#include <util/delay.h>

#include "SwUart.h"


// Timer values for 9600 baud
#define TICKS2COUNT         32  //!< Ticks between two bits.
#define TICKS2WAITONE       32  //!< Wait one bit period.
#define TICKS2WAITONE_HALF  48  //!< Wait one and a half bit period.
#define CSxx                0x03 //!< Clock Slect IOclk/64 prescaler

// Modify for different pre-scaller??????????????????????????????????????????????????????????????
#define INTERRUPT_EXEC_CYCL   9       //!< Cycles to execute interrupt rutine from interrupt.


#define ENABLE_TIMER_INTERRUPT( )       ( TIMSK1 |= ( 1<< OCIE1A ) )
#define DISABLE_TIMER_INTERRUPT( )      ( TIMSK1 &= ~( 1<< OCIE1A ) )
#define CLEAR_TIMER_INTERRUPT( )        ( TIFR1 |= ((1 << OCF1A) ) )

// JWF enable INT0 irq, on ATTiny this is INT0 bit in GIMSK
#define ENABLE_EXTERNAL0_INTERRUPT( )   ( GIMSK |= ( 1<< INT0 ) )
#define DISABLE_EXTERNAL0_INTERRUPT( )  ( GIMSK &= ~( 1<< INT0 ) )
#define CLEAR_EXTERNAL0_INTERRUPT( )    ( GIFR |= (1 << INTF0 ) )

// TX pins
#define TXDDR  DDRA
#define TXPORT PORTA
#define TXPIN  PINA
// RX pins
#define RXDDR  DDRB
#define RXPORT PORTB
#define RXPIN  PINB

#define TX_PIN           PA3                //!< Transmit data pin
#define RX_PIN           PB2                //!< Receive data pin, must be INT0
#define TCCR             TCCR1A             //!< Timer/Counter Control Register
#define TCCR_P           TCCR1B             //!< Timer/Counter Control (Prescaler) Register
#define OCR              OCR1A              //!< Output Compare Register

// JWF on ATTiny replace 'External IRQ registers' with 'General irq registers'
#define EXT_IFR          GIFR               //!< External Interrupt Flag Register
#define EXT_ICR          MCUCR              //!< External Interrupt Control Register

// Macros to modify pin states
#define SET_TX_PIN( )    ( TXPORT |= ( 1 << TX_PIN ) )
#define CLEAR_TX_PIN( )  ( TXPORT &= ~( 1 << TX_PIN ) )
#define GET_RX_PIN( )    ( RXPIN & ( 1 << RX_PIN ) )


typedef enum
{
    IDLE = 0,                                       //!< Idle state, both transmit and receive possible.
    TRANSMIT,                                   //!< Transmitting byte.
    TRANSMIT_STOP_BIT,                          //!< Transmitting stop bit.
    RECEIVE,                                    //!< Receiving byte.
    DATA_PENDING                                //!< Byte received and ready to read.
    
}AsynchronousStates_t;

static volatile AsynchronousStates_t state;     //!< Holds the state of the UART.
static volatile unsigned char SwUartTXData;     //!< Data to be transmitted.
static volatile unsigned char SwUartTXBitCount; //!< TX bit counter.
static volatile unsigned char SwUartRXData;     //!< Storage for received bits.
static volatile unsigned char SwUartRXBitCount; //!< RX bit counter.

volatile uint8_t ttt = 1;


/*! \brief  External interrupt service routine.
 *
 *  The falling edge in the beginning of the start
 *  bit will trig this interrupt. The state will
 *  be changed to RECEIVE, and the timer interrupt
 *  will be set to trig one and a half bit period
 *  from the falling edge. At that instant the
 *  code should sample the first data bit.
 *
 *  \note  initSoftwareUart( void ) must be called in advance.
 */
ISR(EXT_INT0_vect)
{
    state = RECEIVE;                  // Change state
    DISABLE_EXTERNAL0_INTERRUPT( );   // Disable interrupt during the data bits.
 
#if 0
    ttt++;
    if(ttt & 1)
        SET_TX_PIN( );
    else
        CLEAR_TX_PIN();
#endif

    DISABLE_TIMER_INTERRUPT( );       // Disable timer to change its registers.
//    TCCR_P &= ~( 1 << CS01 );         // Reset prescaler counter.
    
    TCNT1 = 0; //INTERRUPT_EXEC_CYCL;      // Clear counter register. Include time to run interrupt rutine.
    
//    TCCR_P |= ( 1 << CS01 );          // Start prescaler clock.
    
    OCR = TICKS2WAITONE_HALF;         // Count one and a half period into the future.
    
    SwUartRXBitCount = 0;             // Clear received bit counter.
    CLEAR_TIMER_INTERRUPT( );         // Clear interrupt bits
    ENABLE_TIMER_INTERRUPT( );        // Enable timer0 interrupt on again

    //////////////////////////// DEBUG
    //SET_TX_PIN( );
    //ttt = 1;
}


/*! \brief  Timer0 interrupt service routine.
 *
 *  Timer0 will ensure that bits are written and
 *  read at the correct instants in time.
 *  The state variable will ensure context
 *  switching between transmit and recieve.
 *  If state should be something else, the
 *  variable is set to IDLE. IDLE is regarded
 *  as a safe state/mode.
 *
 *  \note  initSoftwareUart( void ) must be called in advance.
 */
ISR(TIM1_COMPA_vect) {
   
    switch (state) {
            // Transmit Byte.
        case TRANSMIT:
            // Output the TX buffer.
            if( SwUartTXBitCount < 8 ) {
                if( SwUartTXData & 0x01 ) {           // If the LSB of the TX buffer is 1:
                    SET_TX_PIN();                       // Send a logic 1 on the TX_PIN.
                }
                else {                                // Otherwise:
                    CLEAR_TX_PIN();                     // Send a logic 0 on the TX_PIN.
                }
                SwUartTXData = SwUartTXData >> 1;     // Bitshift the TX buffer and
                SwUartTXBitCount++;                   // increment TX bit counter.
            }
            
            //Send stop bit.
            else {
                SET_TX_PIN();                         // Output a logic 1.
                state = TRANSMIT_STOP_BIT;
            }
            break;
            
            // Go to idle after stop bit was sent.
        case TRANSMIT_STOP_BIT:
            DISABLE_TIMER_INTERRUPT( );           // Stop the timer interrupts.
            state = IDLE;                         // Go back to idle.
            ENABLE_EXTERNAL0_INTERRUPT( );        // Enable reception again.
            break;
            
            //Receive Byte.
        case RECEIVE:
            OCR = TICKS2WAITONE;                  // Count one period after the falling edge is trigged.
#if 0
            ttt++;
            if(ttt & 1)
                SET_TX_PIN( );
            else
                CLEAR_TX_PIN();

#endif

            //Receiving, LSB first.
            if( SwUartRXBitCount < 8 ) {
                SwUartRXBitCount++;
                SwUartRXData = (SwUartRXData>>1);   // Shift due to receiving LSB first.
                if( GET_RX_PIN( ) != 0 ) {
                    SwUartRXData |= 0x80;           // If a logical 1 is read, let the data mirror this.
                }
            }
            
            //Done receiving
            else {
                state = DATA_PENDING;               // Enter DATA_PENDING when one byte is received.
                DISABLE_TIMER_INTERRUPT( );         // Disable this interrupt.
                EXT_IFR |= (1 << INTF0 );           // Reset flag not to enter the ISR one extra time.
                ENABLE_EXTERNAL0_INTERRUPT( );      // Enable interrupt to receive more bytes.
            }
            break;
        
        case DATA_PENDING:
            break;
            // Unknown state.
        default:
            state = IDLE;                           // Error, should not occur. Going to a safe state.
    }
}


/*! \brief  Function to initialize the software UART.
 *
 *  This function will set up pins to transmit and
 *  receive on. Control of Timer0 and External interrupt 0.
 *
 *  \param  void
 *
 *  \retval void
 */
// JWF just butchered this loverly confiurable code to make it work... Must re-do
void initSoftwareUart( void )
{
    // JWF Need to move HW setup to a global app setup routine
    //PORT
    RXPORT |= ( 1 << RX_PIN );       // RX_PIN is input, tri-stated.
    TXDDR |= ( 1 << TX_PIN );        // TX_PIN is output.
    SET_TX_PIN( );                    // Set the TX line to idle state.
    
    // Timer0
    DISABLE_TIMER_INTERRUPT( );
//    TCCR = 0x00;    //Init.
//    TCCR_P = 0x00;    //Init.
//    TCCR |= (1 << WGM01);			// Timer in CTC mode.
//    TCCR_P |= CSxx;                 // Set prescaler.
    TCCR1A = 0b00000000;            // CTC mode
    TCCR1B = 0b00001011;
    
    //External interrupt
    EXT_ICR &= 0b11111100;                   // Init.
    EXT_ICR |= ( 1 << ISC01 );        // Interrupt sense control: falling edge.
    ENABLE_EXTERNAL0_INTERRUPT( );    // Turn external interrupt on.
    
    //Internal State Variable
    state = IDLE;
}



/*! \brief  Send a unsigned char.
 *
 *  This function sends a unsigned char on the TX_PIN
 *  using the timer0 isr.
 *
 *  \note  initSoftwareUart( void ) must be called in advance.
 *
 *  \param  c    unsigned char to transmit.
 *
 *  \retval void
 */
void putChar( const unsigned char c )
{
    while( state != IDLE )
    {
        ;                               // Don't send while busy receiving or transmitting.
    }
    
    state = TRANSMIT;
    DISABLE_EXTERNAL0_INTERRUPT( );  // Disable reception.
    SwUartTXData = c;             // Put byte into TX buffer.
    SwUartTXBitCount = 0;
    
//    TCCR_P &= ~( 1 << CS01 );         // Reset prescaler counter.
    TCNT1 = 0;                        // Clear counter register.
//    TCCR_P |= ( 1 << CS01 );          // CTC mode. Start prescaler clock.
    OCR = TICKS2WAITONE; // JWF added some init
    
    CLEAR_TX_PIN();               // Clear TX line...start of preamble.
    
    CLEAR_TX_PIN( );                   // Clear TX line...start of preamble
    
    ENABLE_TIMER_INTERRUPT( );        // Enable interrupt
}


/*! \brief  Print unsigned char string.
 *
 *  \param  const unsigned char*    Pointer to the string that is to be printed.
 *
 *  \retval void
 */
// JWF must make this an irq driven FIFO
void print_string( const unsigned char *data )
{
    while( *data != '\0' )
    {
        putChar( *data++ );
    }
}


// Returns 0 if no char available
unsigned char get_char(void)
{
    uint8_t ret = 0;

    if( state == DATA_PENDING )
    {
        state = IDLE;
        ret = SwUartRXData;
    }
    return ret;
}

            

