/* Name: main.c
 * Author: James Fletcher
 * Copyright: (c) James Fletcher
 * License: This work may be reproduced, modified, distributed, performed, and displayed for any purpose, but must acknowledge "fab-synth". Copyright is retained and must be preserved. The work is provided as is; no warranty is provided, and users accept all liability.
 */

// Fab synth clock and input module
// Set us up as SPI master and generate clock for the rest of the system
// Also send CV data on the sync pin


#include <avr/io.h>
#include <avr/cpufunc.h>
#include <avr/interrupt.h>
#include <util/delay.h>

#include "SwUart.h"

// IO pin usage
#define CV_SYNC_OUT     _BV(PA7)

#define SPI_DI          _BV(PA6)
#define SPI_DO          _BV(PA5)
#define SPI_SCK         _BV(PA4)

#define PC_SERIAL_OUT   _BV(PA3) // Also has LED on this pin
#define PC_SERIAL_IN    _BV(PB2) // INT0 pin

#define SWITCH_3        _BV(PA2)
#define SWITCH_2        _BV(PA1)
#define SWITCH_1        _BV(PA0)


// USICR setting for bit bang clock
static const uint8_t USICR_CLK1 = 0b00010001; // toggle clock
static const uint8_t USICR_CLK2 = 0b00010011; // toggle clock and shift in a bit


/*
 * SPI frame format for audio and cv data
 * Total of 20 bytes timeslots are:
 *   sync, cv-hi, cv-mid, cv-lo, aud0, ... aud15
 * Sync timeslot is empty, it just contains a falling edge on the sync line signaling the start of a new frame
 * Should add a csumto the frame, although it's unclear what should happen to the audio when data is corrupt!
 */
#define FRAME_SYNC_SLOT     0
#define FRAME_CV0_SLOT      1
#define FRAME_CV1_SLOT      2
#define FRAME_CV2_SLOT      3
#define FRAME_AUDIO_START_SLOT   4
uint8_t txFrame[20];
uint8_t txFrameSlotIndex = 0; // index of time slot in frame

uint8_t tempCount = 0;

/*
 * CV data is currently just a value for the key pressed.  Need to extend to include trigger data and other info.
 */
// cv0: bit 7 - trigger
// cv0: bit 6 - gate
// cv1: unused
// cv2: control voltage value
#define CV0_TRIG 0x20
#define CV0_GATE 0x40

// Other...
volatile uint8_t triggerFrame = 0;   // Next frame is a trigger




// On OC0A we toggle SPI clock pin x8 to send byte
ISR(TIM0_COMPA_vect)
{
    if(txFrameSlotIndex == FRAME_SYNC_SLOT)
    {
        // Generate sync pulse as a start of frame signal
        PORTA &= ~(CV_SYNC_OUT); // sync pin low
        PORTA &= ~(CV_SYNC_OUT); // nop
        PORTA &= ~(CV_SYNC_OUT); // nop
        PORTA &= ~(CV_SYNC_OUT); // nop
        PORTA &= ~(CV_SYNC_OUT); // nop
        PORTA &= ~(CV_SYNC_OUT); // nop
        // End sync pulse
        PORTA |= CV_SYNC_OUT; // sync pin high
    }
    else
    {
        uint8_t byte = txFrame[txFrameSlotIndex];
        
        if((txFrameSlotIndex == FRAME_CV0_SLOT) && triggerFrame)
        {
            byte |= CV0_TRIG;
            triggerFrame = 0;
        }
        
        // Put out data
        USIDR = byte;
        
        // Clock it out onto SPI
        // Slow SPI clock down as I've been seeing some reliability issues
        _NOP();
        _NOP();
        _NOP();
        _NOP();
        USICR = USICR_CLK1;
        _NOP();
        _NOP();
        _NOP();
        _NOP();
        USICR = USICR_CLK2;
        _NOP();
        _NOP();
        _NOP();
        _NOP();
        USICR = USICR_CLK1;
        _NOP();
        _NOP();
        _NOP();
        _NOP();
        USICR = USICR_CLK2;
        _NOP();
        _NOP();
        _NOP();
        _NOP();
        USICR = USICR_CLK1;
        _NOP();
        _NOP();
        _NOP();
        _NOP();
        USICR = USICR_CLK2;
        _NOP();
        _NOP();
        _NOP();
        _NOP();
        USICR = USICR_CLK1;
        _NOP();
        _NOP();
        _NOP();
        _NOP();
        USICR = USICR_CLK2;
        _NOP();
        _NOP();
        _NOP();
        _NOP();
        USICR = USICR_CLK1;
        _NOP();
        _NOP();
        _NOP();
        _NOP();
        USICR = USICR_CLK2;
        _NOP();
        _NOP();
        _NOP();
        _NOP();
        USICR = USICR_CLK1;
        _NOP();
        _NOP();
        _NOP();
        _NOP();
        USICR = USICR_CLK2;
        _NOP();
        _NOP();
        _NOP();
        _NOP();
        USICR = USICR_CLK1;
        _NOP();
        _NOP();
        _NOP();
        _NOP();
        USICR = USICR_CLK2;
        _NOP();
        _NOP();
        _NOP();
        _NOP();
        USICR = USICR_CLK1;
        _NOP();
        _NOP();
        _NOP();
        _NOP();
        USICR = USICR_CLK2;
    }
    
    txFrameSlotIndex++;
    if(txFrameSlotIndex>=20)
        txFrameSlotIndex = 0;
}


// USI counter overflow indicates byte rx
ISR(USI_OVF_vect)
{
    // unused
}


int main(void)
{
    uint8_t cv2_old = 0;
    
    // Fill some data into tx buffer
    txFrame[0] = 0;     // sync - never sent
    txFrame[1] = 0;   // cv0
    txFrame[2] = 0;   // cv1
    txFrame[3] = 0;   // cv2
    txFrame[4] = 1;     // aud
    txFrame[5] = 2;     // aud
    txFrame[6] = 3;     // aud
    txFrame[7] = 4;     // aud
    txFrame[8] = 0;     // aud
    txFrame[9] = 0;     // aud
    txFrame[10] = 0;     // aud
    txFrame[11] = 0;     // aud
    txFrame[12] = 0;     // aud
    txFrame[13] = 0;     // aud
    txFrame[14] = 0;     // aud
    txFrame[15] = 0;     // aud
    txFrame[16] = 0;     // aud
    txFrame[17] = 0;     // aud
    txFrame[18] = 0;     // aud
    txFrame[19] = 0;     // aud
    
    
    // Setup pins required as output
    DDRA = CV_SYNC_OUT | SPI_DO | SPI_SCK | PC_SERIAL_OUT;
    DDRB = 0;
    
    // Enable pull up resistors on switch inputs
    PORTA |= (SWITCH_1 | SWITCH_2 | SWITCH_3);
    
    // Set sync pin high
    PORTA |= CV_SYNC_OUT;
    
    // Enable irq for PC serial comms...
    // For now just use LED on PC_SERIAL_OUT line as debug
    PORTA |= PC_SERIAL_OUT; // LED off
    
    // Enable SPI interface
    USICR = 0b00011010; // no irq, 3 wire master mode, ext clk, pos edge, sw
    USISR = 0b11100000; // clear flags, zero counter

    // TCNT0 Generates an irq for each SPI byte to send
    TCCR0A = 0b00000010; // turn off both OC0A and OC0B, set CTC mode
    TCCR0B = 0b00000010; // set CTC mode, prescale = 1/8
    TCNT0 = 0;
    OCR0A = 64; // 39062.5Hz IRQ to kick off next SPI byte (every 512 CPU clocks)
    TIMSK0 = 0b00000010; // enable OC0A irq
    
    initSoftwareUart();

    sei(); // Global IRQ enable
    
    for(;;)
    {
        uint8_t cv0 = 0, cv1 = 0;
        static uint8_t cv2 = 0;
        static uint8_t midikey = 0;
        
        //print_string( "clock\r\n" );
        unsigned char c=get_char();
#if 0
        if(c != 0)
        {
            print_string( "st: " );
            putChar(c);
            print_string( ":ed\r\n" );
        }
#endif
#if 0
        if(c == 'a')
        {
            cv2 = 0x18;
            //PORTA &= ~(PC_SERIAL_OUT); // LED on
        }
        else if(c == 'b')
        {
            cv2 = 0;
            //PORTA |= PC_SERIAL_OUT; // LED off
        }
#endif
        
        if(c == 0)
        {
            ; // no char sent
        }
        else if(c == 0x01)
        {
            midikey = 0;
            //cv2 = 0; // note off
        }
        else
        {
            midikey = c;
            //cv2 = c; // note on
        }

        cv2 = midikey;
        
        // Debounce and add trigger bit detection
        if((PINA & SWITCH_1) == 0)
            cv2 += 0x10;
        if((PINA & SWITCH_2) == 0)
            cv2 += 0x20;
        if((PINA & SWITCH_3) == 0)
            cv2 += 0x40;
        
        if(cv2 == 0)
        {
            ;//PORTA |= PC_SERIAL_OUT; // LED off
        }
        else
        {
            cv0 |= CV0_GATE;
            //PORTA &= ~(PC_SERIAL_OUT); // LED on
        }

        if(cv2 != cv2_old)
            triggerFrame = 1;
        
        txFrame[FRAME_CV0_SLOT] = cv0;
        txFrame[FRAME_CV1_SLOT] = cv1;
        txFrame[FRAME_CV2_SLOT] = cv2;

        cv2_old = cv2;
    }
    
    return 0;
}
